#include "stdafx.h"
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <string.h>
#include "WinModFileTypes.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

int _read16BITof68000(int h, unsigned short *word)
{
	unsigned short rword;
	int erg;

	erg=_read(h,&rword,2L);
	*word=swapword(rword);
	return(erg);
}

int _read32BITof68000(int h, unsigned long *dword)
{
	unsigned long rdword;
	int erg;

	erg=_read(h,&rdword,4L);
	*dword=swapdword(rdword);
	return(erg);
}

int _readNormName(int h, char *buffer, unsigned long len)
{
	unsigned long i;
	int erg;

	erg=_read(h, buffer, len);
	buffer[len]=0;
	for (i=0; i<len; i++)
	{
		if (buffer[i]>0 && buffer[i]<0x20) buffer[i]=0x20;
		// Translate DOS-ASCII to Windows-ASCII
		if (buffer[i]=='\x84') buffer[i]='';
		if (buffer[i]=='\x94') buffer[i]='';
		if (buffer[i]=='\x81') buffer[i]='';
		if (buffer[i]=='\x8E') buffer[i]='';
		if (buffer[i]=='\x99') buffer[i]='';
		if (buffer[i]=='\x9A') buffer[i]='';
		if (buffer[i]=='\xE1') buffer[i]='';
	}

	return(erg);
}

char ModExtension[4];

void Set_ModType(int h, struct TheModFile &ModClass)
{
	char Kennung[5];

	// Kurz die Kennung lesen:
	*Kennung=0;
	lseek(h,1080L,SEEK_SET);
	_read(h,&Kennung,4L); Kennung[4]=0;
	lseek(h,0L,SEEK_SET);

	// Modtyp erkennen:
	ModClass.nSamples=31;
	if (!strcmp(Kennung,"M.K.") || !strcmp(Kennung,"M!K!"))
	{
		// Es knnte das WOW-Format sein:
		if (!strcmp(ModExtension, "WOW"))
		{
			ModClass.nChannels=8;
			strcpy(ModClass.TrackerName,"Grave Composer");
		}	
		else
		{
			ModClass.nChannels=4;
			strcpy(ModClass.TrackerName,"ProTracker");
		}
	}
	else
	if (!strcmp(Kennung,"FLT4"))
	{
		ModClass.nChannels=4;
		strcpy(ModClass.TrackerName,"StarTrekker");
	}
	else
	if (!strcmp(Kennung,"FLT8"))
	{
		ModClass.nChannels=8;
		strcpy(ModClass.TrackerName,"StarTrekker");
	}
	else
	if (!strcmp(Kennung,"4CHN"))
	{
		ModClass.nChannels=4;
		strcpy(ModClass.TrackerName,"StarTrekker");
	}
	else
	if (!strcmp(Kennung,"6CHN"))
	{
		ModClass.nChannels=6;
		strcpy(ModClass.TrackerName,"FastTracker");
	}
	else
	if (!strcmp(Kennung,"8CHN"))
	{
		ModClass.nChannels=8;
		strcpy(ModClass.TrackerName,"FastTracker");
	}
	else
	if (!strcmp(Kennung,"CD81") || !strcmp(Kennung,"OKTA"))
	{
		ModClass.nChannels=8;
		strcpy(ModClass.TrackerName,"Atari Oktalyzer");
	}
	else
	if (!strcmp(Kennung,"16CH") || !strcmp(Kennung,"16CN"))
	{
		ModClass.nChannels=16;
		strcpy(ModClass.TrackerName,"TakeTracker");
	}
	else
	if (!strcmp(Kennung,"32CH") || !strcmp(Kennung,"32CN"))
	{
		ModClass.nChannels=32;
		strcpy(ModClass.TrackerName,"TakeTracker");
	}	 
	else
	if (!strcmp(Kennung+2,"CH") || !strcmp(Kennung+2,"CN"))
	{
		char Value[3]; Value[0]=Kennung[0]; Value[1]=Kennung[1]; Value[2]=0;
		ModClass.nChannels=atoi(Value);
		sprintf(ModClass.TrackerName,"Unknown Tracker with %s Channels", Value);
	}	 
	else
	{
		*Kennung=0;
		ModClass.nSamples=15;
		ModClass.nChannels=4;
		strcpy(ModClass.TrackerName,"NoiseTracker");
	}
}

void InsertNote(unsigned int PattNum, unsigned int TimeIndex, unsigned int ChannelIndex, unsigned long note, struct TheModFile &ModClass)
{
	struct PatternData *p,*lauf;
	struct PatternData *root;

	root=ModClass.Pattern[PattNum];
	p=new struct PatternData;

	p->TimeIndex=TimeIndex;
	p->ChannelIndex=ChannelIndex;
	
	if (ModClass.nSamples>15)
	{
		p->SampleIndex=(((note&0xF0000000L)>>24) | ((note&0xF000L)>>12))&ModClass.nSamples;
		p->Ton=(note&0x0FFF0000L)>>16;
	}
	else
	{
		p->SampleIndex=((note&0xF000L)>>12)&ModClass.nSamples;
		p->Ton=(note&0xFFFF0000L)>>16;
	}

	int hlp=GetNoteIndex(p->Ton);
	p->noteIndex = hlp%12;
	p->noteOctave = hlp/12;

	p->Effekt=(unsigned char)((note&0xF00)>>8);
	p->Effektop=(unsigned char)(note&0xFF);
	if (p->Effekt==12)
	{
		if (p->Effektop>64) p->Effektop=64;
		/*p->Volume=p->Effektop;
		p->Effekt=0; p->Effektop=0;*/
	}
	
	p->Volume=-1;

	
	p->Next=p->Prev=NULL;

	if (!root) ModClass.Pattern[PattNum]=p;
	else
	{
		lauf=root;
		while (lauf)
		{
			if (!lauf->Next) break;
			if (lauf->Next->TimeIndex>TimeIndex) break;
			lauf=lauf->Next;
		}
		p->Next=lauf->Next;
		p->Prev=lauf;
		if (lauf->Next) lauf->Next->Prev=p;
		lauf->Next=p;
	}
}

unsigned int CalcAnzPattern(int h, struct TheModFile &ModClass, long *Value)
{
	unsigned int i,MaxPatternNumber;
	unsigned long MaxLen;
	unsigned long nBytesPerPattern;

	nBytesPerPattern=64L*4L*ModClass.nChannels;

	MaxLen=150L; // Name+SongLen+CIAA+SongArrangement
	if (*ModClass.ModID) MaxLen+=4L;  // Kennung

	for (i=0; i<ModClass.nSamples; i++)
		MaxLen+=30L+ModClass.Samples[i].Length;

	ModClass.nPattern=(filelength(h)-MaxLen)/nBytesPerPattern;
	*Value=(filelength(h)-MaxLen)%nBytesPerPattern;

	if (*Value) // Fehler lag vor!
	{
		MaxPatternNumber=0;
		for (i=0; i<ModClass.SongLength; i++)
		{
			if (ModClass.SongArrangement[i]>MaxPatternNumber)
				MaxPatternNumber=ModClass.SongArrangement[i];
		}
		MaxPatternNumber++; // Fangen mit 0 an, Anzahl ist aber 1!

		if (MaxPatternNumber>ModClass.nPattern)
		{
			// Ist das Mod zu kurz, MUSS ein gefordertes Pattern zwar gelesen werden, aber
			// es berscheidet sich mit Sampledaten
			(*Value)-=nBytesPerPattern;
			while (MaxPatternNumber>=ModClass.nPattern)
			{
				ModClass.nPattern++;
			}
		}
		else
		{
			// Sinds zu viele Bytes, dann hngen sie meisten hinter den Sampledaten.
			// In dem Fall wrden zu viele Pattern vermutet.
			(*Value)+=(ModClass.nPattern-MaxPatternNumber)*nBytesPerPattern;
			ModClass.nPattern=MaxPatternNumber;
		}
	}

	return(ModClass.nPattern);
}

// In STM_Loader ausgelagert:
extern int Read_Stm(int h, struct TheModFile &ModClass, int *Error, long *ErrorValue);
// In S3M_Loader ausgelagert:
extern int Read_S3m(int h, struct TheModFile &ModClass, int *Error, long *ErrorValue);

int Read_Mod(int h, struct TheModFile &ModClass, int *Error, long *ErrorValue)
{
	char ReadNames[40];
	unsigned long ReadDWord;
	long CalcSamplePos;
	unsigned short ReadWord;
	unsigned char ReadByte;
	unsigned int i, Channel, row, PattNum;

	// Angenommen!
	*Error=LOADERROR_NOERROR;

	ModClass.BPMSpeed = 125; // Standard bei MODS
	ModClass.Tempo    =   6; // STandard bei MODS
	ModClass.usesLinearFrequency = false;	// Standard bei MODS
	
	// Modtyp feststellen
	Set_ModType(h, ModClass);

	// Liest den Songnamen inclusive abschiesendem Null-Byte
	_readNormName(h, ReadNames, 20L); ReadNames[20]=0;
	strcpy(ModClass.SongName,ReadNames);

	// Samples lesen
	ModClass.Samples=new struct SampleData [ModClass.nSamples];
	for (i=0; i<ModClass.nSamples; i++)
	{
		// Samplename
		_readNormName(h, ReadNames, 22L); ReadNames[22]=0;
		strcpy(ModClass.Samples[i].Name, ReadNames);

		// Anzahl Sampldaten
		_read16BITof68000(h, &ReadWord);
		ModClass.Samples[i].Length=((unsigned long)ReadWord)<<1;

		// FineTune:
		_read(h, &ReadByte, 1L);
		if (ReadByte&0x08)
			ModClass.Samples[i].FineTune=(signed char)(0xF0|(ReadByte&0xF));
		else
		if (ReadByte&0x0F)
			ModClass.Samples[i].FineTune=(signed char)(ReadByte&0xF);
		else 
			ModClass.Samples[i].FineTune=0;

		// Lautstrke
		_read(h, &ReadByte, 1L);
		ModClass.Samples[i].Volume=ReadByte;

		// RepeatStart
		_read16BITof68000(h, &ReadWord);
		ModClass.Samples[i].RepeatStart=((unsigned long)ReadWord)<<1;

		// RepeatStop
		_read16BITof68000(h, &ReadWord);
		ModClass.Samples[i].RepeatStop=(((unsigned long)ReadWord)<<1) + ModClass.Samples[i].RepeatStart;
		
		// No Loop with these Values!
		if ((ModClass.Samples[i].RepeatStart==ModClass.Samples[i].RepeatStop) || (!ModClass.Samples[i].RepeatStart && ModClass.Samples[i].RepeatStop==2) || ModClass.Samples[i].RepeatStop==0xFFFF)
		{
			ModClass.Samples[i].RepeatStart=0;
			ModClass.Samples[i].RepeatStop=0;
		}
		else
		// Begrenzen auf Samplelnge, falls nicht korrekt. (sicher ist sicher!)
		if (ModClass.Samples[i].RepeatStop>ModClass.Samples[i].Length) 
			ModClass.Samples[i].RepeatStop=ModClass.Samples[i].Length;

		// Restlichen Werte:
		ModClass.Samples[i].BaseFrequenzy = FineTuneTable[(unsigned)(ModClass.Samples[i].FineTune&0xF)];
		ModClass.Samples[i].BaseTone = getTonToFrequenzy(ModClass.Samples[i].BaseFrequenzy);
	}

	// Songlnge
	_read(h, &ReadByte, 1L);
	if (ReadByte>128) 
	{
		ReadByte=128;
		*Error=LOADERROR_MODCORRUPT;
	}
	ModClass.SongLength=(unsigned int)ReadByte;

	// CIAA
	_read(h, &ReadByte, 1L);
	ModClass.CIAA=(unsigned int)ReadByte;

	// Arrangement
	unsigned int highestPatternToLoad = 0;
	ModClass.SongArrangement=new unsigned long[128];
	for (i=0; i<128; i++)
	{
		_read(h, &ReadByte, 1L);
		if (ReadByte > highestPatternToLoad) highestPatternToLoad = ReadByte;
		ModClass.SongArrangement[i]=ReadByte;
	}

	// Die Kennung
	if (ModClass.nSamples>15)
	{
		_read(h, ReadNames, 4L); ReadNames[4]=0;
		strcpy(ModClass.ModID, ReadNames);
	}
	else *ModClass.ModID=0;

	// Anzahl Pattern bestimmen, die WIRKLICH im Mod sind:
	CalcAnzPattern(h, ModClass, ErrorValue);
	if (*ErrorValue) *Error=LOADERROR_MODSIZE;
/*	else
	{
		// Die verrckte Prfung auf Grave Composer, kann nur bei Fehlerfreien Mods klappen.
		if (ModClass.nChannels == 4 && ((highestPatternToLoad+1)*2) == ModClass.nPattern && !(ModClass.nPattern%2) && !strcmp(ModClass.ModID, "M.K."))
		{
			ModClass.nPattern /= 2;
			ModClass.nChannels *= 2;
			strcpy(ModClass.TrackerName, "Grave Composer");
		}
	}*/

	// Speicher fr alle Pattern reservieren und dann auf NULL initialisieren
	ModClass.Pattern=(struct PatternData **)new (struct PatternData (*[ModClass.nPattern]));
	for (i=0; i<ModClass.nPattern; i++) ModClass.Pattern[i]=NULL;

	if (!strcmp(ModClass.ModID,"FLT8"))
	{
		for (PattNum=0; PattNum<ModClass.nPattern; PattNum++)
		{
			for (row=0; row<64; row++)
			{
				for (Channel=0; Channel<4; Channel++)
				{
					_read32BITof68000(h, &ReadDWord);
					if (ReadDWord) InsertNote(PattNum, row, Channel, ReadDWord, ModClass);
				}
			}
			for (row=0; row<64; row++)
			{
				for (Channel=4; Channel<8; Channel++)
				{
					_read32BITof68000(h, &ReadDWord);
					if (ReadDWord) InsertNote(PattNum, row, Channel, ReadDWord, ModClass);
				}
			}
		}
	}
	else
	{
		for (PattNum=0; PattNum<ModClass.nPattern; PattNum++)
		{
			for (row=0; row<64; row++)
			{
				for (Channel=0; Channel<ModClass.nChannels; Channel++)
				{
					_read32BITof68000(h, &ReadDWord);
					if (ReadDWord) InsertNote(PattNum, row, Channel, ReadDWord, ModClass);
				}
			}
		}
	}

	// Sampledaten:
	// Zu erst in der Datei auf der Position aufsetzen, wo die Daten beginnen.
	for (CalcSamplePos=i=0; i<ModClass.nSamples; i++)
		CalcSamplePos+=ModClass.Samples[i].Length;
	CalcSamplePos=filelength(h)-CalcSamplePos;
	// Nur Position versetzen, wenn Mod zu kurz war (also zu weit gelesen wurde)
	if (CalcSamplePos<tell(h)) lseek(h, CalcSamplePos, SEEK_SET);

	for (i=0; i<ModClass.nSamples; i++)
	{
		unsigned long s, SamLen;

		SamLen=ModClass.Samples[i].Length;
		if (SamLen)
		{
			ModClass.Samples[i].Sample=new signed short[SamLen];
			signed char *TempBuffer=new signed char[SamLen];
			_read(h, TempBuffer, SamLen);

			//ModClass.Samples[i].MaxDiff=0;
			for (s=0; s<SamLen; s++)
			{
				ModClass.Samples[i].Sample[s]=PromModTo16Bit(TempBuffer[s]);
			}

			delete [] TempBuffer;
		} 
		else 
			ModClass.Samples[i].Sample=NULL;
	}
	ModClass.FileType = 0;
	ModClass.BaseVolume=MaxVolValue;
	ModClass.DefaultPanning=0;
	ModClass.panningValuesL = new signed long[ModClass.nChannels];
	ModClass.panningValuesR = new signed long[ModClass.nChannels];
	
	for (i=0; i<ModClass.nChannels; i++)
	{
		if (i%3)
		{
			ModClass.panningValuesR[i]=LMULTI; 
			ModClass.panningValuesL[i]=0;
		}
		else
		{
			ModClass.panningValuesL[i]=LMULTI; 
			ModClass.panningValuesR[i]=0;
		}
	}

	return(1); // Success
}

char *strupper(char *str)
{
	char *c;

	for (c=str; *c; c++) *c=toupper(*c);
	return(str);
}

int Read_ModFiles(char *FileName, struct TheModFile &ModClass, int *Error, long *ErrorValue)
{
	int handle, erg;
	char *extensionStart = strrchr(FileName,'.');

	if (!extensionStart)
	{
		*Error=LOADERROR_UNSUPPORTED;
		*ErrorValue=0;
		return(0);
	}
	strcpy(ModExtension,extensionStart+1);
	strupper(ModExtension);

	handle=_open(FileName, _O_BINARY | _O_RDONLY);
	if (handle==-1) { *Error=LOADERROR_FILENOTFOUND; return(0); }

	// Hier knnen alle erkannten Modfiles mit ihrem Loader aufgerufen werden
	erg=0;
	if (!strcmp(ModExtension,"MOD") || !strcmp(ModExtension,"NST") || !strcmp(ModExtension,"WOW"))
		erg=Read_Mod(handle, ModClass, Error, ErrorValue);
	else
	if (!strcmp(ModExtension,"STM"))
		erg=Read_Stm(handle, ModClass, Error, ErrorValue);
	else
	if (!strcmp(ModExtension,"S3M"))
		erg=Read_S3m(handle, ModClass, Error, ErrorValue);

	// Entweder findet der Loader heraus, da er das Format nicht kann, oder aber, 
	// es wurde gar nicht erst probiert...
	if (!erg)
	{
		*Error=LOADERROR_UNSUPPORTED;
		*ErrorValue=0;
	}

	_close(handle);
	return(erg);
}